/* dvdcpy - A DVD backup tool
 * Copyright (C) 2004-2005 Olivier Rolland <billl@users.sf.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include "config.h"

#include <dvdread/dvd_reader.h>
#include <dvdread/ifo_read.h>

#include <sys/stat.h>

#include <fcntl.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef HAVE_STATVFS
#include <sys/statvfs.h>
#endif

#define CLAMP(x, low, high)  (((x) > (high)) ? (high) : (((x) < (low)) ? (low) : (x)))

unsigned int UDFFindFile (dvd_reader_t *, const char *, unsigned int *);

static unsigned long long total_size;
static unsigned long long current_size;

/*
 * Usage
 */

static void
usage (const char *name)
{
  fprintf (stderr, "usage: %s [-o output] [-t title]* device\n", name);
  fprintf (stderr, "  device    - Specify the DVD device\n");
  fprintf (stderr, "  -o output - Specify the output directory\n");
  fprintf (stderr, "  -t title  - Select the title to copy\n");
  fprintf (stderr, "  -m        - Copy the DVD menu\n");
}

/*
 * Helper functions
 */

static int
make_dir (const char *dir)
{
  struct stat buf;

  if (mkdir (dir, 0755) < 0)
    if (stat (dir, &buf) < 0 || !S_ISDIR (buf.st_mode))
      return -1;

  return 0;
}

static int
make_path (const char *path)
{
  char *dir, *slash;
  int len;

  if (path[0] == '/' && path[1] == '\0')
    return 0;

  slash = dir = strdup (path);

  len = strlen (dir);
  while (len > 1 && dir[--len] == '/')
    dir[len] = '\0';

  while (1)
  {
    while (*slash == '/')
      slash ++;

    slash = strchr (slash, '/');
    if (!slash)
      break;

    *slash = '\0';

    if (make_dir (dir) < 0)
    {
      free (dir);
      return -1;
    }

    *slash++ = '/';
  }

  if (make_dir (dir) < 0)
  {
    free (dir);
    return -1;
  }

  free (dir);

  return 0;
}

static int
make_pathes (const char *path1, ...)
{
  va_list ap;
  char path[FILENAME_MAX];

  strncpy (path, path1, FILENAME_MAX);

  va_start (ap, path1);
  while (path1)
  {
    if (make_path (path) < 0)
    {
      va_end (ap);
      return -1;
    }

    path1 = va_arg (ap, const char *);
    if (path1)
    {
      strncat (path, "/", FILENAME_MAX);
      strncat (path, path1, FILENAME_MAX);
    }
  }
  va_end (ap);

  return 0;
}

static long long
get_space_left (const char *path)
{
#ifdef HAVE_STATVFS
  struct statvfs buf;
#else
  struct statfs buf;
#endif

#ifdef HAVE_STATVFS
  if (statvfs (path, &buf) < 0)
    return -1;
#else
  if (statfs (path, &buf) < 0)
    return -1;
#endif

  return (long long) buf.f_bsize * (long long) buf.f_bavail;
}

static int
dvd_copy_bytes (dvd_file_t *file, unsigned long long size, int fd)
{
  unsigned char *buffer;
  unsigned int copied;

  buffer = (unsigned char *) malloc (size * sizeof (char));
  if (!buffer)
  {
    fprintf (stderr, "Error: Cannot allocate memory\n");
    return -1;
  }

  if (DVDReadBytes(file, buffer, size) != size)
  {
    free (buffer);
    return -1;
  }

  copied = write (fd, buffer, size);

  current_size += copied;
  fprintf (stdout, "\r%llu/%llu bytes written (%llu%%)", current_size, total_size, current_size * 100 / total_size);
  fflush (stdout);

  free (buffer);

  return copied;
}

static int
dvd_copy_blocks (dvd_file_t *file, unsigned long long offset, unsigned long long size, int fd)
{
  unsigned char buffer[512 * DVD_VIDEO_LB_LEN];
  unsigned int start;
  size_t count;

  offset /= DVD_VIDEO_LB_LEN;
  size /= DVD_VIDEO_LB_LEN;

  for (count = 512, start = offset; offset < start + size; offset += count)
  {
    if (offset + count > start + size)
      count = start + size - offset;

    if (DVDReadBlocks (file, offset, count, buffer) != count)
    {
      fprintf (stderr, "Error: Cannot read data\n");
      return -1;
    }

    if (write (fd, buffer, count * DVD_VIDEO_LB_LEN) != count * DVD_VIDEO_LB_LEN)
    {
      fprintf (stderr, "Error: Cannot write data\n");
      return -1;
    }

    current_size += count * DVD_VIDEO_LB_LEN;
    fprintf (stdout, "\r%llu/%llu bytes written (%llu%%)", current_size, total_size, current_size * 100 / total_size);
    fflush (stdout);
  }

  return size * DVD_VIDEO_LB_LEN;
}

/*
 * Get size functions
 */

static unsigned long long
dvd_get_ifo_size (dvd_reader_t *reader, unsigned int vts)
{
  char filename[FILENAME_MAX];
  unsigned int size;

  if (vts == 0)
    strncpy (filename, "/VIDEO_TS/VIDEO_TS.IFO", FILENAME_MAX);
  else
    snprintf (filename, FILENAME_MAX, "/VIDEO_TS/VTS_%02u_0.IFO", vts);

  if (!UDFFindFile (reader, filename, &size))
  {
    fprintf (stderr, "Warning: Cannot find IFO for VTS %u\n", vts);
    return 0;
  }

  return size;
}

static unsigned long long
dvd_get_bup_size (dvd_reader_t *reader, unsigned int vts)
{
  char filename[FILENAME_MAX];
  unsigned int size;

  if (vts == 0)
    strncpy (filename, "/VIDEO_TS/VIDEO_TS.BUP", FILENAME_MAX);
  else
    snprintf (filename, FILENAME_MAX, "/VIDEO_TS/VTS_%02u_0.BUP", vts);

  if (!UDFFindFile (reader, filename, &size))
  {
    fprintf (stderr, "Warning: Cannot find BUP for VTS %u\n", vts);
    return 0;
  }

  return size;
}

static unsigned long long
dvd_get_menu_size (dvd_reader_t *reader, unsigned int vts)
{
  char filename[FILENAME_MAX];
  unsigned int size;

  if (vts == 0)
    strncpy (filename, "/VIDEO_TS/VIDEO_TS.VOB", FILENAME_MAX);
  else
    snprintf (filename, FILENAME_MAX, "/VIDEO_TS/VTS_%02u_0.VOB", vts);

  if (!UDFFindFile (reader, filename, &size))
  {
    fprintf (stderr, "Warning: Cannot find menu for VTS %u\n", vts);
    return 0;
  }

  return size;
}

static long long
dvd_get_vob_size (dvd_reader_t *reader, unsigned int vts)
{
  char filename[FILENAME_MAX];
  unsigned int vob, size;
  unsigned long long fullsize;

  if (vts == 0)
    return 0;

  vob = 1; fullsize = 0;
  while (1)
  {
    snprintf (filename, FILENAME_MAX, "/VIDEO_TS/VTS_%02u_%u.VOB", vts, vob++);
    if (!UDFFindFile (reader, filename, &size))
      break;
    fullsize += size;
  }

  if (vob == 1)
    return -1;

  return fullsize;
}

static long long
dvd_get_vts_size (dvd_reader_t *reader, unsigned int vts)
{
  unsigned long long fullsize;
  long long size;

  fullsize  = dvd_get_ifo_size (reader, vts);
  fullsize += dvd_get_bup_size (reader, vts);
  fullsize += dvd_get_menu_size (reader, vts);

  if (vts > 0)
  {
    if ((size = dvd_get_vob_size (reader, vts)) < 0)
      return -1;
    fullsize += size;
  }

  return fullsize;
}

static long long
dvd_get_vmg_size (dvd_reader_t *reader)
{
  unsigned long long fullsize;

  fullsize  = dvd_get_ifo_size (reader, 0);
  fullsize += dvd_get_bup_size (reader, 0);
  fullsize += dvd_get_menu_size (reader, 0);

  return fullsize;
}

static long long
dvd_get_size (dvd_reader_t *reader, ifo_handle_t *vmg_file, unsigned int vmg, unsigned int *vtss)
{
  long long size = 0, fullsize = 0;
  unsigned int vts;

  if (vmg && (fullsize = dvd_get_vmg_size (reader)) < 0)
    return -1;

  for (vts = 0; vts < vmg_file->vmgi_mat->vmg_nr_of_title_sets; vts ++)
  {
    if (vtss[vts+1])
    {
      if ((size = dvd_get_vts_size (reader, vts + 1)) < 0)
        return -1;
      fullsize += size;
    }
  }

  return fullsize;
}

/*
 * Copy functions
 */

static long long
dvd_copy_ifo (dvd_reader_t *reader, unsigned int vts, const char *output)
{
  char filename[FILENAME_MAX];
  unsigned int size, copied;
  struct stat buf;
  int fd;

  dvd_file_t *file;

  if (vts == 0)
    snprintf (filename, FILENAME_MAX, "%s/VIDEO_TS/VIDEO_TS.IFO", output);
  else
    snprintf (filename, FILENAME_MAX, "%s/VIDEO_TS/VTS_%02u_0.IFO", output, vts);

  if (!UDFFindFile (reader, filename + strlen (output), &size))
  {
    fprintf (stderr, "Warning: Cannot find IFO for VTS %u\n", vts);
    return 0;
  }

  if (stat (filename, &buf) < 0)
    fd = open (filename, O_WRONLY | O_CREAT, 0644);
  else
  {
    if (!S_ISREG (buf.st_mode))
    {
      fprintf (stderr, "Error: %s already exists and is not a regular file\n", filename);
      return -1;
    }
    fd = open (filename, O_WRONLY | O_TRUNC, 0644);
  }

  if (fd < 0)
  {
    fprintf (stderr, "Error: Cannot open %s\n", filename);
    return -1;
  }

  file = DVDOpenFile (reader, vts, DVD_READ_INFO_FILE);
  if (!file)
  {
    fprintf (stderr, "Errot: Cannot open IFO for VTS %u\n", vts);
    close (fd);
    return -1;
  }

  fprintf (stdout, "Copying IFO for VTS %u\n", vts);
  copied = dvd_copy_bytes (file, size, fd);
  fprintf (stdout, "\n");

  DVDCloseFile (file);

  close (fd);

  if (copied != size)
    return -1;

  return size;
}

static long long
dvd_copy_bup (dvd_reader_t *reader, unsigned int vts, const char *output)
{
  char filename[FILENAME_MAX];
  unsigned int size, copied;
  struct stat buf;
  int fd;

  dvd_file_t *file;

  if (vts == 0)
    snprintf (filename, FILENAME_MAX, "%s/VIDEO_TS/VIDEO_TS.BUP", output);
  else
    snprintf (filename, FILENAME_MAX, "%s/VIDEO_TS/VTS_%02u_0.BUP", output, vts);

  if (!UDFFindFile (reader, filename + strlen (output), &size))
  {
    fprintf (stderr, "Warning: Cannot find BUP for VTS %u\n", vts);
    return 0;
  }

  if (stat (filename, &buf) < 0)
    fd = open (filename, O_WRONLY | O_CREAT, 0644);
  else
  {
    if (!S_ISREG (buf.st_mode))
    {
      fprintf (stderr, "Error: %s already exists and is not a regular file\n", filename);
      return -1;
    }
    fd = open (filename, O_WRONLY | O_TRUNC, 0644);
  }

  if (fd < 0)
  {
    fprintf (stderr, "Error: Cannot open %s\n", filename);
    return -1;
  }

  file = DVDOpenFile (reader, vts, DVD_READ_INFO_BACKUP_FILE);
  if (!file)
  {
    fprintf (stderr, "Errot: Cannot open BUP for VTS %u\n", vts);
    close (fd);
    return -1;
  }

  fprintf (stdout, "Copying BUP for VTS %u\n", vts);
  copied = dvd_copy_bytes (file, size, fd);
  fprintf (stdout, "\n");

  DVDCloseFile (file);

  close (fd);

  if (copied != size)
    return -1;

  return size;
}

static long long
dvd_copy_menu (dvd_reader_t *reader, unsigned int vts, const char *output)
{
  char filename[FILENAME_MAX];
  unsigned int size, copied;
  struct stat buf;
  int fd;

  dvd_file_t *file;

  if (vts == 0)
    snprintf (filename, FILENAME_MAX, "%s/VIDEO_TS/VIDEO_TS.VOB", output);
  else
    snprintf (filename, FILENAME_MAX, "%s/VIDEO_TS/VTS_%02u_0.VOB", output, vts);

  if (!UDFFindFile (reader, filename + strlen (output), &size))
  {
    fprintf (stderr, "Warning: Cannot find menu for VTS %u\n", vts);
    return 0;
  }

  if (stat (filename, &buf) < 0)
    fd = open (filename, O_WRONLY | O_CREAT, 0644);
  else
  {
    if (!S_ISREG (buf.st_mode))
    {
      fprintf (stderr, "Error: %s already exists and is not a regular file\n", filename);
      return -1;
    }
    fd = open (filename, O_WRONLY | O_TRUNC, 0644);
  }

  if (fd < 0)
  {
    fprintf (stderr, "Error: Cannot open %s\n", filename);
    return -1;
  }

  file = DVDOpenFile (reader, vts, DVD_READ_MENU_VOBS);
  if (!file)
  {
    fprintf (stderr, "Errot: Cannot open menu for VTS %u\n", vts);
    close (fd);
    return -1;
  }

  fprintf (stdout, "Copying menu for VTS %u\n", vts);
  copied = dvd_copy_blocks (file, 0, size, fd);
  fprintf (stdout, "\n");

  DVDCloseFile (file);

  close (fd);

  if (copied != size)
    return -1;

  return size;
}

static long long
dvd_copy_vob (dvd_reader_t *reader, unsigned int vts, unsigned int vob, unsigned long long offset, const char *output)
{
  char filename[FILENAME_MAX];
  unsigned int size, copied;
  struct stat buf;
  int fd;

  dvd_file_t *file;

  snprintf (filename, FILENAME_MAX, "%s/VIDEO_TS/VTS_%02u_%u.VOB", output, vts, vob);

  if (!UDFFindFile (reader, filename + strlen (output), &size))
    return 0;

  if (stat (filename, &buf) < 0)
    fd = open (filename, O_WRONLY | O_CREAT, 0644);
  else
  {
    if (!S_ISREG (buf.st_mode))
    {
      fprintf (stderr, "Error: %s already exists and is not a regular file\n", filename);
      return -1;
    }
    fd = open (filename, O_WRONLY | O_TRUNC, 0644);
  }

  if (fd < 0)
  {
    fprintf (stderr, "Error: Cannot open %s\n", filename);
    return -1;
  }

  file = DVDOpenFile (reader, vts, DVD_READ_TITLE_VOBS);
  if (!file)
  {
    fprintf (stderr, "Errot: Cannot open VOB %u for VTS %u\n", vob, vts);
    close (fd);
    return -1;
  }

  fprintf (stdout, "Copying VOB %u for VTS %u\n", vob, vts);
  copied = dvd_copy_blocks (file, offset, size, fd);
  fprintf (stdout, "\n");

  DVDCloseFile (file);

  close (fd);

  if (copied != size)
    return -1;

  return size;
}

static int
dvd_copy_vmg (dvd_reader_t *reader, const char *output)
{
  if (dvd_copy_ifo (reader, 0, output) < 0)
    return -1;

  if (dvd_copy_bup (reader, 0, output) < 0)
    return -1;

  if (dvd_copy_menu (reader, 0, output) < 0)
    return -1;

  return 0;
}

static int
dvd_copy_vts (dvd_reader_t *reader, unsigned int vts, const char *output)
{
  unsigned int vob;
  long long offset, size;

  if (dvd_copy_ifo (reader, vts, output) < 0)
    return -1;

  if (dvd_copy_bup (reader, vts, output) < 0)
    return -1;

  if (dvd_copy_menu (reader, vts, output) < 0)
    return -1;

  if (vts > 0)
  {
    vob = 0; offset = 0;
    while ((size = dvd_copy_vob (reader, vts, ++vob, offset, output)) != -1)
    {
      if (size == 0)
        break;
      offset += size;
    }
  }

  return 0;
}

static int
dvd_copy (dvd_reader_t *reader, ifo_handle_t *vmg_file, unsigned int vmg, unsigned int *vtss, const char *output)
{
  unsigned int vts;

  if (vmg && dvd_copy_vmg (reader, output) < 0)
    return -1;

  for (vts = 0; vts < vmg_file->vmgi_mat->vmg_nr_of_title_sets; vts++)
    if (vtss[vts+1] && dvd_copy_vts (reader, vts+1, output) < 0)
      return -1;

  return 0;
}

int
main (int argc, char *argv[])
{
  unsigned int i, vmg, title, titles[100], vtss[100];

  dvd_reader_t *reader;
  ifo_handle_t *vmg_file;
  char *output;
  int c;

  if (argc < 2)
  {
    usage (argv[0]);
    return EXIT_FAILURE;
  }

  title = vmg = 0;
  output = "backup";

  memset (titles, 0, 100 * sizeof (int));

  while ((c = getopt (argc, argv, "hmo:t:")) != -1)
  {
    switch (c)
    {
      case 'h':
        usage (argv[0]);
        return EXIT_SUCCESS;
        break;
      case 'm':
        vmg = 1;
        break;
      case 'o':
        output = optarg;
        break;
      case 't':
        title = atoi (optarg);
        if (title < 1 || title > 99)
        {
          fprintf (stderr, "Error: title must be greater than 0 and lesser than 100\n");
          return EXIT_FAILURE; 
        }
        titles[title - 1] = 1;
        break;
      default:
        usage (argv[0]);
        return EXIT_FAILURE;
        break;
    }
  }

  if (optind != argc - 1)
  {
    usage (argv[0]);
    return EXIT_FAILURE;
  }

  reader = DVDOpen (argv[optind]);
  if (!reader)
  {
    fprintf (stderr, "Error: Cannot open DVD device\n");
    return EXIT_FAILURE; 
  }

  vmg_file = ifoOpen (reader, 0);
  if (!vmg_file)
  {
    fprintf (stderr, "Error: Cannot open VMG info\n");
    DVDClose (reader);
    return EXIT_FAILURE;
  }

  if (!title)
    vmg = 1;

  memset (vtss, 0, vmg_file->tt_srpt->nr_of_srpts * sizeof (int));
  for (i = 0; i < vmg_file->tt_srpt->nr_of_srpts; i++)
    vtss[vmg_file->tt_srpt->title[i].title_set_nr] |= !title || titles[i];

  if (make_pathes (output, "VIDEO_TS", NULL) < 0)
  {
    fprintf (stderr, "Error: Cannot create DVD structure\n");
    ifoClose (vmg_file);
    DVDClose (reader);
    return EXIT_FAILURE;
  }

  if ((total_size = dvd_get_size (reader, vmg_file, vmg, vtss)) < 0)
  {
    ifoClose (vmg_file);
    DVDClose (reader);
    return EXIT_FAILURE;
  }

  if (total_size > get_space_left (output))
  {
    fprintf (stderr, "Error: Not enough space left on device (%lld needed)\n", total_size);
    ifoClose (vmg_file);
    DVDClose (reader);
    return EXIT_FAILURE;
  }

  if ((dvd_copy (reader, vmg_file, vmg, vtss, output)) < 0)
  {
    ifoClose (vmg_file);
    DVDClose (reader);
    return EXIT_FAILURE;
  }

  ifoClose (vmg_file);

  DVDClose (reader);

  return EXIT_SUCCESS;
}

