/* OGMSpawn - A library to spawn processes
 * Copyright (C) 2004-2005 Olivier Rolland <billl@esil.univ-mrs.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include "ogmsp-pipeline.h"
#include "ogmsp-enums.h"

#define OGMSP_PIPELINE_GET_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((o), OGMSP_TYPE_PIPELINE, OGMSpPipelinePriv))

struct _OGMSpPipelinePriv
{
  gdouble fraction;
};

static gint ogmsp_pipeline_run    (OGMSpSpawn     *spawn);
static void ogmsp_pipeline_add    (OGMSpContainer *container, 
                                   OGMSpSpawn     *child);
static void ogmsp_pipeline_remove (OGMSpContainer *container, 
                                   OGMSpSpawn     *child);

G_DEFINE_TYPE (OGMSpPipeline, ogmsp_pipeline, OGMSP_TYPE_LIST)

static void
ogmsp_pipeline_class_init (OGMSpPipelineClass *klass)
{
  OGMSpSpawnClass *spawn_class;
  OGMSpContainerClass *container_class;

  spawn_class = OGMSP_SPAWN_CLASS (klass);
  container_class = OGMSP_CONTAINER_CLASS (klass);

  spawn_class->run = ogmsp_pipeline_run;

  container_class->add = ogmsp_pipeline_add;
  container_class->remove = ogmsp_pipeline_remove;

  g_type_class_add_private (klass, sizeof (OGMSpPipelinePriv));
}

static void
ogmsp_pipeline_init (OGMSpPipeline *pipeline)
{
  pipeline->priv = OGMSP_PIPELINE_GET_PRIVATE (pipeline);
}

static gint
ogmsp_pipeline_run (OGMSpSpawn *spawn)
{
  OGMSpPipeline *pipeline;
  GList *children, *child;

  GError *tmp_error = NULL;
  gint result = OGMSP_RESULT_COMPLETED;

  pipeline = OGMSP_PIPELINE (spawn);

  children = ogmsp_list_get_children (OGMSP_LIST (spawn));
  pipeline->priv->fraction = 0.0;

  for (child = children; child; child = child->next)
  {
    if (result != OGMSP_RESULT_COMPLETED)
      break;

    ogmsp_spawn_set_async (OGMSP_SPAWN (child->data), child->next != NULL);
    result = ogmsp_spawn_run (OGMSP_SPAWN (child->data), &tmp_error);

    if (result == OGMSP_RESULT_ERROR && tmp_error)
      ogmsp_spawn_propagate_error (spawn, tmp_error);
  }
  g_list_free (children);

  return result;
}

static void
ogmsp_pipeline_child_progress (OGMSpPipeline *pipeline, gdouble fraction, OGMSpSpawn *child)
{
  if (fraction > pipeline->priv->fraction)
  {
    pipeline->priv->fraction = fraction;
    g_signal_emit_by_name (pipeline, "progress", fraction);
  }
}

static void
ogmsp_pipeline_add (OGMSpContainer *container, OGMSpSpawn *child)
{
  guint handler;

  handler = g_signal_connect_swapped (child, "progress", 
      G_CALLBACK (ogmsp_pipeline_child_progress), container);
  g_object_set_data (G_OBJECT (child), "__child_progress_handler__", 
      GUINT_TO_POINTER (handler));

  OGMSP_CONTAINER_CLASS (ogmsp_pipeline_parent_class)->add (container, child);
}

static void
ogmsp_pipeline_remove (OGMSpContainer *container, OGMSpSpawn *child)
{
  gpointer handler;

  handler = g_object_get_data (G_OBJECT (child), "__child_progress_handler__");
  if (handler)
    g_signal_handler_disconnect (container, GPOINTER_TO_UINT (handler));

  OGMSP_CONTAINER_CLASS (ogmsp_pipeline_parent_class)->remove (container, child);
}

OGMSpSpawn *
ogmsp_pipeline_new (void)
{
  return g_object_new (OGMSP_TYPE_PIPELINE, NULL);
}

