/* OGMSpawn - A library to spawn processes
 * Copyright (C) 2004-2005 Olivier Rolland <billl@esil.univ-mrs.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include "ogmsp-spawn.h"
#include "ogmsp-marshal.h"
#include "ogmsp-enums.h"

#define OGMSP_SPAWN_GET_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((o), OGMSP_TYPE_SPAWN, OGMSpSpawnPriv))

struct _OGMSpSpawnPriv
{
  gint result;
  gint priority;
  gboolean async;
  GError *error;
  OGMSpSpawn *parent;
};

enum
{
  RUN,
  CANCEL,
  PROGRESS,
  LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 };

static void ogmsp_spawn_dispose (GObject *gobject);

GQuark
ogmsp_spawn_error_quark (void)
{
  static GQuark quark = 0;

  if (quark == 0)
    quark = g_quark_from_static_string ("ogmsp-spawn-error-quark");

  return quark;
}

G_DEFINE_ABSTRACT_TYPE (OGMSpSpawn, ogmsp_spawn, G_TYPE_OBJECT)

static void
ogmsp_spawn_class_init (OGMSpSpawnClass *klass)
{
  GObjectClass *gobject_class;

  gobject_class = G_OBJECT_CLASS (klass);

  gobject_class->dispose = ogmsp_spawn_dispose;

  signals[RUN] = g_signal_new ("run", 
      G_TYPE_FROM_CLASS (klass), G_SIGNAL_RUN_LAST, 
      G_STRUCT_OFFSET (OGMSpSpawnClass, run), 
      NULL, NULL, ogmsp_cclosure_marshal_INT__VOID, 
      G_TYPE_INT, 0);

  signals[CANCEL] = g_signal_new ("cancel", 
      G_TYPE_FROM_CLASS (klass), G_SIGNAL_RUN_FIRST, 
      G_STRUCT_OFFSET (OGMSpSpawnClass, cancel), 
      NULL, NULL, g_cclosure_marshal_VOID__VOID, 
      G_TYPE_NONE, 0);

  signals[PROGRESS] = g_signal_new ("progress", 
      G_TYPE_FROM_CLASS (klass), G_SIGNAL_RUN_FIRST, 
      G_STRUCT_OFFSET (OGMSpSpawnClass, progress), 
      NULL, NULL, g_cclosure_marshal_VOID__DOUBLE, 
      G_TYPE_NONE, 1, G_TYPE_DOUBLE);

  g_type_class_add_private (klass, sizeof (OGMSpSpawnPriv));
}

static void
ogmsp_spawn_init (OGMSpSpawn *spawn)
{
  spawn->priv = OGMSP_SPAWN_GET_PRIVATE (spawn);
  spawn->priv->priority = G_MININT;
}

static void
ogmsp_spawn_dispose (GObject *gobject)
{
  OGMSpSpawn *spawn;

  spawn = OGMSP_SPAWN (gobject);

  /*
   * Cancel ?
   */

  G_OBJECT_CLASS (ogmsp_spawn_parent_class)->dispose (gobject);
}

gint
ogmsp_spawn_run (OGMSpSpawn *spawn, GError **error)
{
  g_return_val_if_fail (OGMSP_IS_SPAWN (spawn), OGMSP_RESULT_ERROR);
  g_return_val_if_fail (error == NULL || *error == NULL, OGMSP_RESULT_ERROR);

  spawn->priv->result = OGMSP_RESULT_ERROR;
  spawn->priv->error = NULL;

  g_signal_emit (spawn, signals[RUN], 0, &spawn->priv->result);

  if (spawn->priv->result == OGMSP_RESULT_ERROR && spawn->priv->error)
    g_propagate_error (error, spawn->priv->error);

  return spawn->priv->result;
}

void
ogmsp_spawn_cancel (OGMSpSpawn *spawn)
{
  g_return_if_fail (OGMSP_IS_SPAWN (spawn));

  if (spawn->priv->result != OGMSP_RESULT_CANCEL)
  {
#ifdef OGMSP_DEBUG
    g_printerr ("Canceling %s\n", G_OBJECT_TYPE_NAME (spawn));
#endif

    spawn->priv->result = OGMSP_RESULT_CANCEL;

    g_signal_emit (spawn, signals[CANCEL], 0);
/*
    if (spawn->priv->parent)
      ogmsp_spawn_cancel (spawn->priv->parent);
*/
  }
}

void
ogmsp_spawn_set_async (OGMSpSpawn *spawn, gboolean async)
{
  g_return_if_fail (OGMSP_IS_SPAWN (spawn));

  spawn->priv->async = async;
}

gboolean
ogmsp_spawn_get_async (OGMSpSpawn *spawn)
{
  g_return_val_if_fail (OGMSP_IS_SPAWN (spawn), FALSE);

  return spawn->priv->async;
}

OGMSpSpawn *
ogmsp_spawn_get_parent (OGMSpSpawn *spawn)
{
  g_return_val_if_fail (OGMSP_IS_SPAWN (spawn), NULL);

  return spawn->priv->parent;
}

void
ogmsp_spawn_set_parent (OGMSpSpawn *spawn, OGMSpSpawn *parent)
{
  g_return_if_fail (OGMSP_IS_SPAWN (spawn));
  g_return_if_fail (parent == NULL || OGMSP_IS_SPAWN (parent));

  spawn->priv->parent = parent;
}

void
ogmsp_spawn_propagate_error (OGMSpSpawn *spawn, GError *error)
{
  g_return_if_fail (OGMSP_IS_SPAWN (spawn));

  if (error)
    g_propagate_error (&spawn->priv->error, error);
}

void
ogmsp_spawn_set_priority (OGMSpSpawn *spawn, gint priority)
{
  g_return_if_fail (OGMSP_IS_SPAWN (spawn));

  spawn->priv->priority = CLAMP (priority, -20, 19);
}

gint
ogmsp_spawn_get_priority (OGMSpSpawn *spawn)
{
  g_return_val_if_fail (OGMSP_IS_SPAWN (spawn), G_MININT);

  if (spawn->priv->priority != G_MININT)
    return spawn->priv->priority;

  if (spawn->priv->parent)
    return ogmsp_spawn_get_priority (spawn->priv->parent);

  return 0;
}

